@@ -10,7 +10,16 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+from io import BytesIO
+import tarfile
+import time
+import types
+
+from oslo_serialization import jsonutils as json
 from oslo_utils import encodeutils
+import six
+from tempest.lib.common.utils import data_utils
+from tempest.lib.common.utils import test_utils
 from tempest.lib import decorators
 
 from zun_tempest_plugin.tests.tempest.api import clients
@@ -42,6 +51,9 @@ class TestContainer(base.BaseZunTest):
         cls.images_client = cls.os_primary.images_client
         cls.ports_client = cls.os_primary.ports_client
         cls.sgs_client = cls.os_primary.sgs_client
+        cls.networks_client = cls.os_primary.neutron_client
+        cls.subnets_client = cls.os_primary.subnets_client
+        cls.vol_client = cls.os_primary.vol_client
 
     @classmethod
     def resource_setup(cls):
@@ -124,27 +136,28 @@ class TestContainer(base.BaseZunTest):
 
     @decorators.idempotent_id('c32f93e3-da88-4c13-be38-25d2e662a28e')
     def test_run_container_with_image_driver_glance(self):
-        image = None
-        try:
-            docker_base_url = self._get_docker_url()
-            self.docker_client.pull_image(
-                'cirros', docker_auth_url=docker_base_url)
-            image_data = self.docker_client.get_image(
-                'cirros', docker_base_url)
-            image = self.images_client.create_image(
-                name='cirros', disk_format='raw', container_format='docker')
-            self.images_client.store_image_file(image['id'], image_data)
-            # delete the local image that was previously pulled down
-            self.docker_client.delete_image('cirros', docker_base_url)
+        docker_base_url = self._get_docker_url()
+        self.docker_client.pull_image(
+            'cirros', docker_auth_url=docker_base_url)
+        image_data = self.docker_client.get_image(
+            'cirros', docker_base_url)
+        if isinstance(image_data, types.GeneratorType):
+            # NOTE(kiennt): In Docker-py 3.1.0, get_image
+            #               returns generator [1]. These lines
+            #               makes image_data readable.
+            # [1] https://bugs.launchpad.net/zun/+bug/1753080
+            image_data = six.b('').join(image_data)
+            image_data = six.BytesIO(image_data)
 
-            _, model = self._run_container(
-                image='cirros', image_driver='glance')
-        finally:
-            if image:
-                try:
-                    self.images_client.delete_image(image['id'])
-                except Exception:
-                    pass
+        image = self.images_client.create_image(
+            name='cirros', disk_format='raw', container_format='docker')
+        self.addCleanup(self.images_client.delete_image, image['id'])
+        self.images_client.store_image_file(image['id'], image_data)
+        # delete the local image that was previously pulled down
+        self.docker_client.delete_image('cirros', docker_base_url)
+
+        _, model = self._run_container(
+            image='cirros', image_driver='glance')
 
     @decorators.idempotent_id('b70bedbc-5ba2-400c-8f5f-0cf05ca17151')
     def test_run_container_with_environment(self):
@@ -170,6 +183,20 @@ class TestContainer(base.BaseZunTest):
         self.assertTrue('key1=label1', labels)
         self.assertTrue('key2=label2', labels)
 
+    @decorators.idempotent_id('8fc7fec1-e1a2-3f65-a5a6-dba425c1607c')
+    def test_run_container_with_port(self):
+        project_id = self.container_client.tenant_id
+        networks = self.networks_client.list_networks()['networks']
+        for network in networks:
+            if network['project_id'] == project_id:
+                network_id = network['id']
+
+        port = self.ports_client.create_port(network_id=network_id,
+                                             name='testport')
+        _, model = self._run_container(nets=[{'port': 'testport'}])
+        self.assertTrue('testport', model)
+        self.addCleanup(self.ports_client.delete_port, port['port']['id'])
+
     @decorators.idempotent_id('9fc7fec0-e1a9-4f65-a5a6-dba425c1607c')
     def test_run_container_with_restart_policy(self):
         _, model = self._run_container(restart_policy={
@@ -205,13 +232,119 @@ class TestContainer(base.BaseZunTest):
 
     @decorators.idempotent_id('f181eeda-a9d1-4b2e-9746-d6634ca81e2f')
     def test_run_container_with_security_groups(self):
-        sg_name = 'test_sg'
-        self.sgs_client.create_security_group(name=sg_name)
+        sg_name = data_utils.rand_name('test_sg')
+        sg = self.sgs_client.create_security_group(name=sg_name)
+        self.addCleanup(self.sgs_client.delete_security_group,
+                        sg['security_group']['id'])
         _, model = self._run_container(security_groups=[sg_name])
         sgs = self._get_all_security_groups(model)
         self.assertEqual(1, len(sgs))
         self.assertEqual(sg_name, sgs[0])
 
+    @decorators.idempotent_id('f55dbe0d-3e8a-4798-9267-c9b13361e721')
+    def test_run_container_with_network(self):
+        """Test container run with the given network
+
+        This test does the following:
+        1. Create a network and its subnet
+        2. Verity the created network and its subnet.
+        3. Run a container with this network.
+        4. Verify container's addresses is in subnet's cidr.
+        """
+        test_net = self.networks_client.create_network(
+            name='test_net')['network']
+        self.assertEqual(test_net['name'], 'test_net')
+        test_subnet = self.subnets_client.create_subnet(
+            name='test_subnet', network_id=test_net['id'], ip_version=4,
+            cidr='10.1.0.0/24')['subnet']
+        self.assertEqual(test_subnet['name'], 'test_subnet')
+        self.assertEqual(test_subnet['cidr'], '10.1.0.0/24')
+        _, model = self._run_container(nets=[{'network': test_net['id']}])
+        self.assertEqual(1, len(model.addresses))
+        subnet_id = list(model.addresses.values())[0][0]['subnet_id']
+        addr = list(model.addresses.values())[0][0]['addr']
+        self.assertEqual(subnet_id, test_subnet['id'])
+        self.assertIn('10.1.0', addr)
+        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
+                        self.networks_client.delete_network, test_net['id'])
+
+    @decorators.idempotent_id('2bc86759-ffca-4b3d-bf25-4cf260a67704')
+    def test_run_container_with_shared_network(self):
+        """Test container run with the given shared network
+
+        This test does the following:
+        1. Create a network and its subnet (In admin tenant)
+        2. Verity the created network and its subnet.
+        3. Run a container with this network.
+        4. Verify container's addresses is in subnet's cidr.
+        """
+        test_net = self.os_admin.neutron_client.create_network(
+            name='test_net', shared=True)['network']
+        self.assertEqual(test_net['name'], 'test_net')
+        test_subnet = self.os_admin.subnets_client.create_subnet(
+            name='test_subnet', network_id=test_net['id'], ip_version=4,
+            cidr='10.1.0.0/24')['subnet']
+        self.assertEqual(test_subnet['name'], 'test_subnet')
+        self.assertEqual(test_subnet['cidr'], '10.1.0.0/24')
+        _, model = self._run_container(nets=[{'network': test_net['id']}])
+        self.assertEqual(1, len(model.addresses))
+        subnet_id = list(model.addresses.values())[0][0]['subnet_id']
+        addr = list(model.addresses.values())[0][0]['addr']
+        self.assertEqual(subnet_id, test_subnet['id'])
+        self.assertIn('10.1.0', addr)
+        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
+                        self.os_admin.neutron_client.delete_network,
+                        test_net['id'])
+
+    @decorators.idempotent_id('7a947d75-ab23-439a-bd94-f6e219f716a9')
+    def test_run_container_with_cinder_volumes(self):
+        """Tests the following:
+
+        1. Create a volume in cinder.
+        2. Run a container with the volume mounted into a path in file system.
+        3. Execute a command in the container to write some data to the volume.
+        4. Delete the container (the volume is untouched).
+        5. Create a new container with the same volume mounted.
+        6. Execute a command in the container to read data from the volume.
+        7. Assert the data read from the volume is the same as the data
+           written in before.
+        """
+        # create a volume in cinder
+        volume = self.vol_client.create_volume(
+            name=data_utils.rand_name(), size=1)['volume']
+        volume_id = volume['id']
+
+        # create a container with the volume
+        container_path = '/data'
+        container_file = '/data/testfile'
+        _, model = self._run_container(mounts=[{
+            'source': volume_id, 'destination': container_path}])
+        volume = self.vol_client.show_volume(volume_id)['volume']
+        self.assertEqual('in-use', volume['status'])
+        # write data into the volume
+        resp, _ = self.container_client.exec_container(
+            model.uuid,
+            command="/bin/sh -c 'echo hello > %s'" % container_file)
+        self.assertEqual(200, resp.status)
+        # delete the container
+        resp, _ = self.container_client.delete_container(
+            model.uuid, params={'stop': True})
+        self.assertEqual(204, resp.status)
+        self.container_client.ensure_container_deleted(model.uuid)
+        volume = self.vol_client.show_volume(volume_id)['volume']
+        self.assertEqual('available', volume['status'])
+
+        # create another container with the same volume
+        _, model = self._run_container(mounts=[{
+            'source': volume_id, 'destination': container_path}])
+        volume = self.vol_client.show_volume(volume_id)['volume']
+        self.assertEqual('in-use', volume['status'])
+        # read data from the volume
+        resp, body = self.container_client.exec_container(
+            model.uuid, command='cat %s' % container_file)
+        self.assertEqual(200, resp.status)
+        self.assertTrue('hello' in encodeutils.safe_decode(body))
+
     @decorators.idempotent_id('c3f02fa0-fdfb-49fc-95e2-6e4dc982f9be')
     def test_commit_container(self):
         """Test container snapshot
@@ -345,6 +478,7 @@ class TestContainer(base.BaseZunTest):
         self.assertEqual(200, resp.status)
         self.assertTrue('hello' in encodeutils.safe_decode(body))
 
+    @decorators.skip_because(bug="1764923")
     @decorators.idempotent_id('d383f359-3ebd-40ef-9dc5-d36922790230')
     def test_update_container(self):
         _, model = self._run_container(cpu=0.1, memory=100)
@@ -365,15 +499,18 @@ class TestContainer(base.BaseZunTest):
                                                      docker_base_url)
         self._assert_resource_constraints(container, cpu=0.2, memory=200)
 
+    @decorators.skip_because(bug="1759459")
     @decorators.idempotent_id('b218bea7-f19b-499f-9819-c7021ffc59f4')
     def test_rename_container(self):
-        _, model = self._run_container(name='container1')
-        self.assertEqual('container1', model.name)
-        gen_model = datagen.container_rename_data(name='container2')
+        container1_name = data_utils.rand_name('container1')
+        _, model = self._run_container(name=container1_name)
+        self.assertEqual(container1_name, model.name)
+        container2_name = data_utils.rand_name('container2')
+        gen_model = datagen.container_rename_data(name=container2_name)
         resp, model = self.container_client.rename_container(model.uuid,
                                                              gen_model)
         self.assertEqual(200, resp.status)
-        self.assertEqual('container2', model.name)
+        self.assertEqual(container2_name, model.name)
 
     @decorators.idempotent_id('142b7716-0b21-41ed-b47d-a42fba75636b')
     def test_top_container(self):
@@ -397,14 +534,15 @@ class TestContainer(base.BaseZunTest):
         self.assertTrue('BLOCK I/O(B)' in encodeutils.safe_decode(body))
 
     @decorators.idempotent_id('b3b9cf17-82ad-4c1b-a4af-8210a778a33e')
-    def test_add_sg_to_container(self):
+    def test_add_and_remove_sg_to_container(self):
         _, model = self._run_container()
         sgs = self._get_all_security_groups(model)
         self.assertEqual(1, len(sgs))
         self.assertEqual('default', sgs[0])
-
-        sg_name = 'test_add_sg'
-        self.sgs_client.create_security_group(name=sg_name)
+        sg_name = data_utils.rand_name('test_add_sg')
+        sg = self.sgs_client.create_security_group(name=sg_name)
+        self.addCleanup(self.sgs_client.delete_security_group,
+                        sg['security_group']['id'])
         gen_model = datagen.container_add_sg_data(name=sg_name)
         resp, body = self.container_client.add_security_group(
             model.uuid, gen_model)
@@ -421,6 +559,19 @@ class TestContainer(base.BaseZunTest):
 
         utils.wait_for_condition(assert_security_group_is_added)
 
+        gen_model = datagen.container_remove_sg_data(name=sg_name)
+        resp1, body = self.container_client.remove_security_group(
+            model.uuid, gen_model)
+        self.assertEqual(202, resp1.status)
+
+        def assert_security_group_is_removed():
+            sgs = self._get_all_security_groups(model)
+            if sg_name not in sgs:
+                return True
+            else:
+                return False
+        utils.wait_for_condition(assert_security_group_is_removed)
+
     def _assert_resource_constraints(self, container, cpu=None, memory=None):
         if cpu is not None:
             cpu_quota = container.get('HostConfig').get('CpuQuota')
@@ -519,8 +670,8 @@ class TestContainer(base.BaseZunTest):
     @decorators.idempotent_id('dcb0dddb-7f0f-43f6-b82a-0cae13938bd6')
     def test_detach_and_attach_network_to_container(self):
         _, model = self._run_container()
-        self.assertEqual(1, len(model.addresses))
 
+        self.assertEqual(1, len(model.addresses))
         network = list(model.addresses.keys())[0]
         resp, body = self.container_client.network_detach(
             model.uuid, params={'network': network})
@@ -529,6 +680,43 @@ class TestContainer(base.BaseZunTest):
             model.uuid, params={'network': network})
         self._ensure_network_attached(model, network)
 
+    @decorators.idempotent_id('037d800c-2262-4e15-90cd-95292b5ef958')
+    def test_put_and_get_archive_from_container(self):
+        _, model = self._run_container()
+        self.assertEqual(1, len(model.addresses))
+
+        # Create a simple tarstream
+        file_content = 'Random text'
+
+        tarstream = BytesIO()
+        with tarfile.open(fileobj=tarstream, mode='w') as tar:
+            encoded_file_content = file_content.encode()
+            tarinfo = tarfile.TarInfo(name='test.txt')
+            tarinfo.size = len(encoded_file_content)
+            tarinfo.mtime = time.time()
+            tar.addfile(tarinfo, BytesIO(encoded_file_content))
+
+        # We're at the end of the tarstream, go back to the beginning
+        tarstream.seek(0)
+
+        req_body = json.dump_as_bytes({'data': tarstream.getvalue()})
+        resp, _ = self.container_client.put_archive(
+            model.uuid, params={'path': '/tmp'}, body=req_body)
+        self.assertEqual(200, resp.status)
+        resp, body = self.container_client.get_archive(
+            model.uuid, params={'path': '/tmp/test.txt'})
+        self.assertEqual(200, resp.status)
+
+        # Get content
+        body = json.loads(body)
+        tardata = BytesIO(body['data'].encode())
+        with tarfile.open(fileobj=tardata, mode='r') as tar:
+            untar_content = tar.extractfile('test.txt').read()
+
+        self.assertEqual(file_content, untar_content.decode())
+        self.assertEqual(body['stat']['name'], tarinfo.name)
+        self.assertEqual(body['stat']['size'], tarinfo.size)
+
     def _ensure_network_detached(self, container, network):
         def is_network_detached():
             _, model = self.container_client.get_container(container.uuid)
